//======================================================================================================
// LIB.H
// Copyright 2022 Everett Kaser
// All rights reserved
//======================================================================================================

typedef	unsigned char		BYTE;
typedef unsigned short		WORD;
typedef	unsigned long		DWORD;
typedef unsigned long long	QWORD;

typedef int					BOOL;
typedef DWORD				HANDLE;
typedef unsigned long		size_t;
typedef BYTE *				va_list;

#define NULL	((void *)0)
#define	FALSE	0
#define	TRUE	1

//======================================================================================================
//======================================================================================================
#define	DEBUG	0

//======================================================================================================
//======================================================================================================

#if DEBUG
	#define	SM(s)	SystemMessage(s)
	#define	SMS()	SystemMessage(ScratBuf)
	#define	SP()	sprintf(ScratBuf,
	void DumpStack(char *head);
	void WriteDebug(char *p);
#else
	#define	SM(s)	Junk Junk Junk SMjunk
	#define	SMS()	Junk Junk Junk SMSjunk
	#define	SP()	Junk Junk Junk SPjunk
#endif

typedef struct {
	DWORD	x, y, w, h, showcmd;
} WNDPOS;

//======================================================================================================
// To get the relative offset of a structure element within the structure:
#define offsetof(s,m)	(size_t)&(((s *)0)->m)

// Rounds sizeof(n) up to the next multiple of sizeof(int)
#define _INTSIZEOF(n)	( (sizeof(n) + sizeof(int) - 1) & ~(sizeof(int) - 1) )

//======================================================================================================
// For variable length argument lists, as in printf()
//======================================================================================================
#define va_start(ap,v)	( ap = (va_list)&v + _INTSIZEOF(v) )
#define va_arg(ap,t)	( *(t *)((ap += _INTSIZEOF(t)) - _INTSIZEOF(t)) )
#define va_end(ap)		( ap = (va_list)0 )

//======================================================================================================
// test macros
//======================================================================================================
#define max(a,b)    (((a) > (b)) ? (a) : (b))
#define min(a,b)    (((a) < (b)) ? (a) : (b))
#define	abs(a)		(((a) < 0)? (-(a)) : (a))

#define	BitsB(n)	BitCounts[(n) & 0x0FF]
#define	BitsW(n)	(BitsB(n) + BitsB((n)>>8))
#define	BitsD(n)	(BitsW(n) + BitsW((n)>>16))
#define	BitsQ(n)	(BitsD(n) + BitsD((n)>>32))

//======================================================================================================
// CHARACTER DEFINES
// NOTE: The 'alpha' defines/functions ONLY operate on ASCII characters, NOT UTF8 characters!
//======================================================================================================
extern BYTE _ctype[];

// bit masks character types

#define _UPPER        0x1       // upper case letter
#define _LOWER        0x2       // lower case letter
#define _DIGIT        0x4       // digit[0-9]
#define _SPACE        0x8       // tab, carriage return, newline, vertical tab or form feed
#define _PUNCT       0x10       // punctuation character
#define _CTL	     0x20       // control character
#define _BLANK       0x40       // space char
#define _HEX         0x80       // hexadecimal digit

// char macros

#define isalpha(c)      ( (_ctype)[c] & (_UPPER|_LOWER) )
#define isupper(c)      ( (_ctype)[c] & _UPPER )
#define islower(c)      ( (_ctype)[c] & _LOWER )
#define isdigit(c)      ( (_ctype)[c] & _DIGIT )
#define isxdigit(c)     ( (_ctype)[c] & _HEX )
#define isspace(c)      ( (_ctype)[c] & _SPACE )
#define ispunct(c)      ( (_ctype)[c] & _PUNCT )
#define isalnum(c)      ( (_ctype)[c] & (_UPPER|_LOWER|_DIGIT) )
#define isprint(c)      ( (_ctype)[c] & (_BLANK|_PUNCT|_UPPER|_LOWER|_DIGIT) )
#define isgraph(c)      ( (_ctype)[c] & (_PUNCT|_UPPER|_LOWER|_DIGIT) )
#define iscntrl(c)      ( (_ctype)[c] & _CONTROL )
#define isascii(c)      ( (unsigned)(c) < 0x80 )

#define isalunder(c)	(isalpha(c) || ((c) == '_'))
#define isalnumunder(c)	(isalnum(c) || ((c) == '_'))

//======================================================================================================
// FILE I/O DEFINES
//======================================================================================================

// O_RDBIN is read-only and O_WRBIN is write/read of pre-existing file.
// O_WRBINNEW creates a new file or truncates a pre-existing file and opens
//		it for write/read.  All read/write modes are binary, no text translation.
#define	O_RDBIN			0x08000
#define	O_WRBIN			0x08002
#define	O_WRBINNEW		0x08302

// Seek start points
#define SEEK_SET    0
#define SEEK_CUR    1
#define SEEK_END    2

// maximum length of filepaths
#define FILENAME_MAX    260

//======================================================================================================
// STRINGS and HELP
//======================================================================================================
#define MAXKBUF				65536
#define MAXKBLOCKS			32
#define	MAXKSTRINGS			1024

#define	MAXHELPLEN			0x000C0000	// max length of buffer to hold expanded/parsed help.txt
#define	MAXHELPPAGES		128

typedef struct {
	long	x;
	long	y;
	DWORD	flags;
	long	len;
	char	*p;
} HELPOBJ;
typedef HELPOBJ* HELPOBJPTR;

typedef struct {
	DWORD	back;
	DWORD	text;
	DWORD	frame;
} TABLECOLOR;

//======================================================================================================
// EVENT MESSAGES
//======================================================================================================
#define	KMSG_INVALID_MSG		0	/// SHOULD never be sent
#define	KMSG_BEGIN_STARTUP		1	/// program is JUST starting up, window not yet created
#define	KMSG_END_STARTUP		2	/// program has finished starting up, window created, ready for normal run
#define	KMSG_QUERY_QUIT			3	/// asking if it's OK to quite the program (return 0=no, 1=yes)
#define	KMSG_SHUTDOWN			4	/// program is about to shutdown
#define	KMSG_FOCUS				5	/// KINT window got or lost input focus
#define	KMSG_SCREEN_CHANGE		6	/// screen has changed, resize/redraw EVERYTHING
#define	KMSG_WINDOW_HIDDEN		7	/// window has been minimized (arg1==1) or un-minimized (arg1==0)
#define	KMSG_WINDOW_MOVED		8	/// window has been moved
#define	KMSG_WINDOW_SIZED		9	/// window has been resized
#define	KMSG_MUSIC_DONE			10	/// music has finished playing
#define	KMSG_TIMER				11	/// a timer has elapsed
#define	KMSG_KEY_DOWN			12	/// a key has been pressed
#define	KMSG_KEY_UP				13	/// a key has been released
#define	KMSG_CHAR				14	/// a character has been typed (NOTE: a single key press may cause KEY_DOWN, CHAR, KEY_UP)
#define	KMSG_POINTER_MOVE		15	/// a pointer device (mouse, finger?) has moved (mouse button clicks come through as KEY_DOWN/KEY_UP)

//=================================================================
typedef	struct {
	long	x, y;
} POINT;

typedef struct {
	long	x, y, w, h;
} RECT;

typedef struct {
	long	w, h;
	RECT	clip;
} INFOBM;

//=======================================================================
// Pre-declare KINTLIB.C functions
//=======================================================================

void DrawLine(DWORD hBM, long x1, long y1, long x2, long y2, DWORD clr);
int sprintf(char *buf, char *format, ...);

//=======================================================================
// Pre-declare KINTLIB.C SYS functions
//=======================================================================

DWORD _Sysfopen(DWORD cmd, char *fname, long flags);
long _Sysfeof(DWORD cmd, DWORD fh);
long _Sysfread(DWORD cmd, DWORD fh, void *buf, long count);
long _Sysfwrite(DWORD cmd, DWORD fh, void *buf, long count);
DWORD _Sysfseek(DWORD cmd, DWORD fh, long offset, DWORD origin);
void _Sysfremove(DWORD cmd, char *fname);
long _Sysmkdir(DWORD cmd, char *fname);
void _Sysrmdir(DWORD cmd, char *fname);
void _Sysfclose(DWORD cmd, DWORD fh);
DWORD _Sysfsize(DWORD cmd, char *fname);
QWORD _Sysftime(DWORD cmd, char *fname);
DWORD _SysFindFirst(DWORD cmd, char *wildname, char *buf, long *dirflag, long *appdata);
long _SysFindNext(DWORD cmd, char *buf, long *dirflag, long *appdata);
void _SysFindClose(DWORD cmd);
long _Sysfwrite64(DWORD cmd, DWORD fh, void *buf, long count);
void _Sysfrename(DWORD cmd, char *fname, char *tname);

void _SysSetTimer(DWORD cmd, DWORD val);
DWORD _SysTimeMilli(DWORD cmd);
void _SysTimeStr(DWORD cmd, char *buf);

long _SysPrintFmt(DWORD cmd, char *dbuf, char *fbuf, long size, QWORD str);
void _Sysstrcpy(DWORD cmd, char *s, char *z);
void _Sysstrcat(DWORD cmd, char *s, char *z);
long _Sysstrlen(DWORD cmd, char *s);
long _Sysstrcmp(DWORD cmd, char *s, char *z);
void _Syscharupper(DWORD cmd, char *s);
void _Sysmemcpy(DWORD cmd, BYTE *d, BYTE *s, long len);
void _Sysmemmove(DWORD cmd, BYTE *d, BYTE *s, long len);
void _Sysmemset(DWORD cmd, BYTE *d, DWORD v, long len);
long _Sysmemcmp(DWORD cmd, BYTE *s, BYTE *z, long len);
void _Sysdbgout(DWORD cmd, char *s);

DWORD _SysCreateBM(DWORD cmd, long w, long h);
void _SysDeleteBM(DWORD cmd, DWORD hBM);
DWORD _SysLoadBM(DWORD cmd, char *fname);
DWORD _SysSaveBM(DWORD cmd, DWORD hBM, char *fname, long format);

void _SysDrawLinesBM(DWORD cmd, DWORD hBM, POINT *pPoints, long numPoints, DWORD clr);
void _SysDrawRectBM(DWORD cmd, DWORD hBM, long lf, long tp, long rt, long bt, DWORD fill, DWORD edge);
void _SysDrawGradientRectBM(DWORD cmd, DWORD hBM, long xo, long yo, long wo, long ho, WORD hue, WORD satmin, WORD satmax, WORD lightmin, WORD lightmax);
void _SysDrawOvalBM(DWORD cmd, DWORD hBM, long x1, long y1, long w, long h, DWORD fill, DWORD edge);
void _SysDrawBM(DWORD cmd, DWORD hBMD, long xd, long yd, DWORD hBMS, long xs, long ys, long w, long h, DWORD a, long rep2x, long rep2y);
void _SysStretchBM(DWORD cmd, DWORD hBMD, long xd, long yd, long wd, long hd, DWORD hBMS, long xs, long ys, long ws, long hs, long rep2x, long rep2y);
void _SysGetInfoBM(DWORD cmd, DWORD hBM, INFOBM *pIBM);
void _SysClipBM(DWORD cmd, DWORD hBM, long l, long t, long r, long b, RECT *pRECT);
void _SysColorizeBM(DWORD cmd, DWORD hBM, long x, long y, long w, long h, long hue, long sat, long lit, long cont);
DWORD _SysDrawTextBM(DWORD cmd, DWORD hBM, char *pText, long x, long y, long w, long h, DWORD flags, DWORD bclr, DWORD fclr);
QWORD _SysSetAlphaBM(DWORD cmd, DWORD hBM, long x, long y, long w, long h, QWORD clr, DWORD alpha);
DWORD _SysGetPixelBM(DWORD cmd, DWORD hBM, long x, long y);
void _SysSetPixelBM(DWORD cmd, DWORD hBM, long x, long y, DWORD clr);

QWORD _SysSetLineStyle(DWORD cmd, QWORD ls);

DWORD _SysRGB2HSL(DWORD cmd, DWORD argb);
DWORD _SysHSL2RGB(DWORD cmd, DWORD ahsl);

void _SysPlaySound(DWORD cmd, char *fname);
void _SysPlayMusic(DWORD cmd, char *fname);
void _SysStopMusic(DWORD cmd);
DWORD _SysIsMusic(DWORD cmd);

void _SysSystemMessage(DWORD cmd, char *pMSG);
long _SysGetEOL(DWORD cmd, char *buf, long maxbuflen);
QWORD _SysGetScreenSize(DWORD cmd);
DWORD _SysGetKeyMsg(DWORD cmd, DWORD *buf);
DWORD _SysGetPtrMsg(DWORD cmd, DWORD *buf);
void _SysReadClipboard(DWORD cmd, char *buf, DWORD buflen);
void _SysWriteClipboard(DWORD cmd, char *buf, DWORD len);
void _SysCapRelMouse(DWORD cmd, DWORD cap);
void _SysGetWndPos(DWORD cmd, WNDPOS *pDW);
void _SysSetWndPos(DWORD cmd, WNDPOS *pDW);
QWORD _SysGetViewSize(DWORD cmd);
void _SysSetPgmTitle(DWORD cmd, char *pBuf);
void _SysGetAppDataPath(DWORD cmd, char *pBuf);
void _SysGetPgmPath(DWORD cmd, char *pBuf);
DWORD _SysGetOSversion(DWORD cmd, char *pBuf);
void _SysGetCtl(DWORD cmd, char *pBuf);
QWORD _SysGetMaxScreen(DWORD cmd);
long _SysGetKeyStatus(DWORD cmd, DWORD key);

BOOL _SysiniRead(DWORD cmd, char *fname);
BOOL _SysiniWrite(DWORD cmd, char *fname);
void _SysiniClose(DWORD cmd);
BOOL _SysiniFindSec(DWORD cmd, char *pSecName);
BOOL _SysiniFindVar(DWORD cmd, char *pVarName);
BOOL _SysiniGetNum(DWORD cmd, DWORD *pNum);
long _SysiniGetStr(DWORD cmd, char *pStr, long maxlen);
BOOL _SysiniSetSec(DWORD cmd, char *pSecName);
BOOL _SysiniSetVar(DWORD cmd, char *pVarName, char *pVarVal);
void _SysiniDelSec(DWORD cmd);
BOOL _SysiniDelVar(DWORD cmd, char *pVarName);
long _SysiniNextSec(DWORD cmd, char *pBuf, long maxlen);
void _SysiniDelAll(DWORD cmd);
BOOL _SysiniIsEndVar(DWORD cmd);
BOOL _SysiniAddVarNum(DWORD cmd, char *pVarName, DWORD val, DWORD format);
BOOL _SysiniAddVarStr(DWORD cmd, char *pVarName, char *pStr);
BOOL _SysiniNextVar(DWORD cmd, char *pBuf, long maxlen);

//======================================================================================================
// Major SYS function groupings
//======================================================================================================
#define	SYS_FILE	0
#define	SYS_PRINT	1
#define	SYS_TIME	2
#define	SYS_STRMEM	3
#define	SYS_WINDOW	4
#define	SYS_GRAPH	5
#define	SYS_SOUND	6
#define	SYS_SYSTEM	7
#define	SYS_CFG		8
#define	SYS_HEAP	9

#define	A_READONLY	1
#define	A_HIDDEN	2
#define	A_SYSTEM	4
#define	A_DIR		16
#define	A_ARCHIVE	32

//==========================================
// SYS_FILE
#define	SYS_FOPEN		((SYS_FILE<<8)+0)
#define	SYS_FEOF		((SYS_FILE<<8)+1)
#define	SYS_FREAD		((SYS_FILE<<8)+2)
#define	SYS_FWRITE		((SYS_FILE<<8)+3)
#define	SYS_FSEEK		((SYS_FILE<<8)+4)
#define	SYS_FREMOVE		((SYS_FILE<<8)+5)
#define	SYS_MKDIR		((SYS_FILE<<8)+6)
#define	SYS_RMDIR		((SYS_FILE<<8)+7)
#define	SYS_FCLOSE		((SYS_FILE<<8)+8)
#define	SYS_FSIZE		((SYS_FILE<<8)+9)
#define	SYS_FTIME		((SYS_FILE<<8)+10)
#define	SYS_FINDFIRST	((SYS_FILE<<8)+11)
#define	SYS_FINDNEXT	((SYS_FILE<<8)+12)
#define	SYS_FINDCLOSE	((SYS_FILE<<8)+13)
#define	SYS_FWRITE64	((SYS_FILE<<8)+14)
#define	SYS_FRENAME		((SYS_FILE<<8)+15)

//==========================================
// SYS_TIME
#define	SYS_SETTIMER	((SYS_TIME<<8)+0)
#define	SYS_TIMEMILLI	((SYS_TIME<<8)+1)
#define	SYS_TIMESTR		((SYS_TIME<<8)+2)

//==========================================
// SYS_STRMEM
#define	SYS_PRTFMT	((SYS_STRMEM<<8)+0)
#define	SYS_STRCPY	((SYS_STRMEM<<8)+1)
#define	SYS_STRCAT	((SYS_STRMEM<<8)+2)
#define	SYS_STRLEN	((SYS_STRMEM<<8)+3)
#define	SYS_STRCMP	((SYS_STRMEM<<8)+4)
#define	SYS_CHARUPPER ((SYS_STRMEM<<8)+5)

#define	SYS_MEMCPY	((SYS_STRMEM<<8)+10)
#define	SYS_MEMMOVE	((SYS_STRMEM<<8)+11)
#define	SYS_MEMSET	((SYS_STRMEM<<8)+12)
#define	SYS_MEMCMP	((SYS_STRMEM<<8)+13)

#define	SYS_DBGOUT	((SYS_STRMEM<<8)+14)

//======================================================================================================
// SYS_GRAPH
// NOTE: the "+n" in these macros MUST match the order in the KINT.EXE
#define	SYS_LINES			((SYS_GRAPH<<8)+0)
#define	SYS_RECT			((SYS_GRAPH<<8)+1)
#define	SYS_LOADBM			((SYS_GRAPH<<8)+2)
#define	SYS_DRAWBM			((SYS_GRAPH<<8)+3)
#define	SYS_DELBM			((SYS_GRAPH<<8)+4)
#define	SYS_SETALPHA		((SYS_GRAPH<<8)+5)
#define	SYS_GETPIXEL		((SYS_GRAPH<<8)+6)
#define	SYS_SETPIXEL		((SYS_GRAPH<<8)+7)
#define	SYS_STRETCHBM		((SYS_GRAPH<<8)+8)
#define	SYS_SAVEBM			((SYS_GRAPH<<8)+9)
#define	SYS_RGB2HSL			((SYS_GRAPH<<8)+10)
#define	SYS_HSL2RGB			((SYS_GRAPH<<8)+11)
#define	SYS_GETINFOBM		((SYS_GRAPH<<8)+12)
#define	SYS_CLIPBM			((SYS_GRAPH<<8)+13)
#define	SYS_SETLINESTYLE	((SYS_GRAPH<<8)+14)
#define	SYS_COLORIZEBM		((SYS_GRAPH<<8)+15)
#define	SYS_DRAWTEXTBM		((SYS_GRAPH<<8)+16)
#define	SYS_DRAWGRADRECTBM	((SYS_GRAPH<<8)+17)
#define	SYS_DRAWOVALBM		((SYS_GRAPH<<8)+18)
#define	SYS_CREATEBM		((SYS_GRAPH<<8)+19)

//==========================================
// SYS_SOUND
#define	SYS_PLAYSOUND		((SYS_SOUND<<8)+0)
#define	SYS_PLAYMUSIC		((SYS_SOUND<<8)+1)
#define	SYS_STOPMUSIC		((SYS_SOUND<<8)+2)
#define	SYS_ISMUSIC			((SYS_SOUND<<8)+3)

//==========================================
// SYS_SYSTEM
#define	SYS_SYSMSG			((SYS_SYSTEM<<8)+0)
#define	SYS_GETEOL			((SYS_SYSTEM<<8)+1)
#define	SYS_GETSCREENSIZE	((SYS_SYSTEM<<8)+2)
#define	SYS_GETKEYMSG		((SYS_SYSTEM<<8)+3)
#define	SYS_GETPTRMSG		((SYS_SYSTEM<<8)+4)
#define	SYS_READCLIPBOARD	((SYS_SYSTEM<<8)+5)
#define	SYS_WRITECLIPBOARD	((SYS_SYSTEM<<8)+6)
#define	SYS_CAPTUREMOUSE	((SYS_SYSTEM<<8)+7)
#define	SYS_GETWNDPOS		((SYS_SYSTEM<<8)+8)
#define	SYS_SETWNDPOS		((SYS_SYSTEM<<8)+9)
#define	SYS_GETVIEWSIZE		((SYS_SYSTEM<<8)+10)
#define	SYS_SETPGMTITLE		((SYS_SYSTEM<<8)+11)
#define	SYS_GETAPPDATAPATH	((SYS_SYSTEM<<8)+12)
#define	SYS_GETPGMPATH		((SYS_SYSTEM<<8)+13)
#define	SYS_GETOSVERSION	((SYS_SYSTEM<<8)+14)
#define	SYS_GETCTL			((SYS_SYSTEM<<8)+15)
#define	SYS_SETCURSOR		((SYS_SYSTEM<<8)+16)
#define	SYS_DOHTML			((SYS_SYSTEM<<8)+17)
#define	SYS_GETMAXSCREEN	((SYS_SYSTEM<<8)+18)
#define	SYS_GETKEYSTATUS	((SYS_SYSTEM<<8)+19)

//==========================================
// SYS_CFG
#define	SYS_INIREAD			((SYS_CFG<<8)+0)	/// OPENS and READS the .ini file into memory
#define	SYS_INIWRITE		((SYS_CFG<<8)+1)	/// WRITES the current INI buffer to the file, but keeps the buffer for further activity (i.e., just flushes it to storage)
#define	SYS_INICLOSE		((SYS_CFG<<8)+2)	/// free the .ini memory resources (losing any changes unless you've already done an INIWRITE)
#define	SYS_INIFINDSEC		((SYS_CFG<<8)+3)	/// find an already-existing section
#define	SYS_INIFINDVAR		((SYS_CFG<<8)+4)	/// try to find the specified variable in the current section, return TRUE if successful (with the data pointer set to that variable, ready to read/write)
#define	SYS_INIGETNUM		((SYS_CFG<<8)+5)	/// get a number from the current variable at the current position in the variable
#define	SYS_INIGETSTR		((SYS_CFG<<8)+6)	/// get a quoted string from the current variable at the current position in the variable
#define	SYS_INISETSEC		((SYS_CFG<<8)+7)	/// sets "current section" to the named section (creating it if it doesn't exist)
#define	SYS_INISETVAR		((SYS_CFG<<8)+8)	/// creates a new, or entirely replaces an old, variable in the current section with a new set of pre-formatted values
#define	SYS_INIDELSEC		((SYS_CFG<<8)+9)	/// delete the current section and all its variables and data
#define	SYS_INIDELVAR		((SYS_CFG<<8)+10)	/// delete the named variable and its value(s)
#define	SYS_ININEXTSEC		((SYS_CFG<<8)+11)	/// move to NEXT section, return its name
#define	SYS_INIDELALL		((SYS_CFG<<8)+12)	/// delete ENTIRE .ini buffer contents, but leave buffer allocated and ready to go
#define	SYS_INIENDVAR		((SYS_CFG<<8)+13)	/// return TRUE if no more data before End-Of-Line of current variable
#define	SYS_INIADDVARNUM	((SYS_CFG<<8)+14)	/// add a numeric value to the end of the variable
#define	SYS_INIADDVARSTR	((SYS_CFG<<8)+15)	/// add a quoted string to the end of the variable
#define	SYS_ININEXTVAR		((SYS_CFG<<8)+16)	/// returns the ENTIRE next variable line

//==========================================
// SYS_HEAP
#define	SYS_HEAPALLOC		((SYS_HEAP<<8)+0)
#define	SYS_HEAPFREE		((SYS_HEAP<<8)+1)
//#define	SYS_HEAPLOCK		((SYS_HEAP<<8)+2)
//#define	SYS_HEAPUNLOCK		((SYS_HEAP<<8)+3)
#define	SYS_HEAPPTR			((SYS_HEAP<<8)+4)

//===========================================
#define	CURSOR_NORMAL		0
#define	CURSOR_WAIT			1
#define	CURSOR_LINK			2

//=======================================================================
// Define/map easy-to-use function forms into ones that have the SYS 'cmd' added to the front
//=======================================================================
#define	fopen(fname, flags)			_Sysfopen(SYS_FOPEN, fname, flags)
#define	feof(fh)					_Sysfeof(SYS_FEOF, fh)
#define	fread(fh, buf, count)		_Sysfread(SYS_FREAD, fh, buf, count)
#define	fwrite(fh, buf, count)		_Sysfwrite(SYS_FWRITE, fh, buf, count)
#define	fseek(fh, off, origin)		_Sysfseek(SYS_FSEEK, fh, off, origin)
#define	fremove(fname)				_Sysfremove(SYS_FREMOVE, fname)
#define	mkdir(fname)				_Sysmkdir(SYS_MKDIR, fname)
#define	rmdir(fname)				_Sysrmdir(SYS_RMDIR, fname)
#define	fclose(fh)					_Sysfclose(SYS_FCLOSE, fh)
#define	fsize(fname)				_Sysfsize(SYS_FSIZE, fname)
#define	ftime(fname)				_Sysftime(SYS_FTIME, fname)
#define	FindFirst(wildname, buf, dirflg, appdata)	_SysFindFirst(SYS_FINDFIRST, wildname, buf, dirflg, appdata)
#define	FindNext(buf, dirflg, appdata)	_SysFindNext(SYS_FINDNEXT, buf, dirflg, appdata)
#define	FindClose()					_SysFindClose(SYS_FINDCLOSE)
#define	fwrite64(fh, buf, count)	_Sysfwrite64(SYS_FWRITE64, fh, buf, count)
#define	frename(fname, tname)		_Sysfrename(SYS_FRENAME, fname, tname)

#define	SetTimer(val)				_SysSetTimer(SYS_SETTIMER, val)
#define	TimeMilli()					_SysTimeMilli(SYS_TIMEMILLI)
#define	TimeStr(buf)				_SysTimeStr(SYS_TIMESTR, buf)

#define	Print(dbuf, fbuf, size, val)				_SysPrintFmt(SYS_PRTFMT, dbuf, fbuf, size, val)
#define	strcpy(s, z)								_Sysstrcpy(SYS_STRCPY, s, z)
#define	strcat(s, z)								_Sysstrcat(SYS_STRCAT, s, z)
#define	strlen(s)									_Sysstrlen(SYS_STRLEN, s)
#define	strcmp(s, z)								_Sysstrcmp(SYS_STRCMP, s, z)
#define	charupper(s)								_Syscharupper(SYS_CHARUPPER, s)
#define	memcpy(d, s, len)							_Sysmemcpy(SYS_MEMCPY, d, s, len)
#define	memmove(d, s, len)							_Sysmemmove(SYS_MEMMOVE, d, s, len)
#define	memset(d, v, len)							_Sysmemset(SYS_MEMSET, d, v, len)
#define	memcmp(s, z, len)							_Sysmemcmp(SYS_MEMCMP, s, z, len)
#define	DBGOUT(s)									_Sysdbgout(SYS_DBGOUT, s)

#define	CreateBM(w, h)								_SysCreateBm(SYS_CREATEBM, w, h)
#define	DeleteBM(hBM)								_SysDeleteBM(SYS_DELBM, hBM)
#define	LoadBM(fname)								_SysLoadBM(SYS_LOADBM, fname)
#define	SaveBM(hBM, fname, format)					_SysSaveBM(SYS_SAVEBM, hBM, fname, format)
#define	DrawLinesBM(hBM, p, n, clr)					_SysDrawLinesBM(SYS_LINES, hBM, p, n, clr)
#define	DrawRectBM(hBM, x, y, w, h, fill, edge)		_SysDrawRectBM(SYS_RECT, hBM, x, y, w, h, fill, edge)
#define	DrawGradientRectBM(hBM, xo, yo, wo, ho, hue, satmin, satmax, lightmin, lightmax)	_SysDrawGradientRectBM(SYS_DRAWGRADRECTBM, hBM, xo, yo, wo, ho, hue, satmin, satmax, lightmin, lightmax)
#define	DrawOvalBM(hBM, x1, y1, w, h, fill, edge)	_SysDrawOvalBM(SYS_DRAWOVALBM, hBM, x1, y1, w, h, fill, edge)
#define	DrawBM(hBMD, xd, yd, hBMS, xs, ys, w, h, a)	_SysDrawBM(SYS_DRAWBM, hBMD, xd, yd, hBMS, xs, ys, w, h, a, 0, 0)
#define	RepDrawBM(hBMD, xd, yd, hBMS, xs, ys, w, h, a, rep2x, rep2y)	_SysDrawBM(SYS_DRAWBM, hBMD, xd, yd, hBMS, xs, ys, w, h, a, rep2x, rep2y)
#define	StretchBM(hBMD, xd, yd, wd, hd, hBMS, xs, ys, ws, hs)	_SysStretchBM(SYS_STRETCHBM, hBMD, xd, yd, wd, hd, hBMS, xs, ys, ws, hs, 0, 0)
#define	RepStretchBM(hBMD, xd, yd, wd, hd, hBMS, xs, ys, ws, hs, tx, ty)	_SysStretchBM(SYS_STRETCHBM, hBMD, xd, yd, wd, hd, hBMS, xs, ys, ws, hs, tx, ty)
#define	GetInfoBM(hBM, pIBM)						_SysGetInfoBM(SYS_GETINFOBM, hBM, pIBM)
#define	ClipBM(hBM, l,t,r,b, pRECT)					_SysClipBM(SYS_CLIPBM, hBM, l, t, r, b, pRECT)
#define	ClipRectBM(hBM, r, rr)						_SysClipBM(SYS_CLIPBM, hBM, r.x, r.y, r.w, r.h, rr)
#define	ColorizeBM(hBM, x, y, w, h, hue, sat, lit, cont)	_SysColorizeBM(SYS_COLORIZEBM, hBM, x, y, w, h, hue, sat, lit, cont)
#define DrawTextBM(hBM, pText, x, y, w, h, flags, bclr, fclr)	_SysDrawTextBM(SYS_DRAWTEXTBM, hBM, pText, x, y, w, h, flags, bclr, fclr)
#define	SetAlphaBM(hBM, x, y, w, h, clr, alpha)		_SysSetAlphaBM(SYS_SETALPHA, hBM, x, y, w, h, clr, alpha)
#define	GetPixelBM(hBM, x, y)						_SysGetPixelBM(SYS_GETPIXEL, hBM, x, y)
#define	SetPixelBM(hBM, x, y, clr)					_SysSetPixelBM(SYS_SETPIXEL, hBM, x, y, clr)
#define	SetLineStyle(ls)							_SysSetLineStyle(SYS_SETLINESTYLE, ls)
#define	RGB2HSL(argb)								_SysRGB2HSL(SYS_RGB2HSL, argb)
#define	HSL2RGB(ahsl)								_SysHSL2RGB(SYS_HSL2RGB, ahsl)

#define	PlaySound(s)								_SysPlaySound(SYS_PLAYSOUND, s)
#define	PlayMusic(s)								_SysPlayMusic(SYS_PLAYMUSIC, s)
#define	StopMusic()									_SysStopMusic(SYS_STOPMUSIC)
#define	IsMusic()									_SysIsMusic(SYS_ISMUSIC)

#define	SystemMessage(pMSG)							_SysSystemMessage(SYS_SYSMSG, pMSG)
#define	GetEOL(buf, maxbuflen)						_SysGetEOL(SYS_GETEOL, buf, maxbuflen)
#define	GetScreenSize()								_SysGetScreenSize(SYS_GETSCREENSIZE)
#define	GetKeyMsg(buf)								_SysGetKeyMsg(SYS_GETKEYMSG, buf)
#define	GetPtrMsg(buf)								_SysGetPtrMsg(SYS_GETPTRMSG, buf)
#define ReadClipboard(buf, buflen)					_SysReadClipboard(SYS_READCLIPBOARD, buf, buflen)
#define	WriteClipboard(buf, buflen)					_SysWriteClipboard(SYS_WRITECLIPBOARD, buf, buflen)
#define	CapRelMouse(cap)							_SysCapRelMouse(SYS_CAPTUREMOUSE, cap)
#define	GetWndPos(pDW)								_SysGetWndPos(SYS_GETWNDPOS, pDW)
#define	SetWndPos(pDW)								_SysSetWndPos(SYS_SETWNDPOS, pDW)
#define	GetViewSize()								_SysGetViewSize(SYS_GETVIEWSIZE)
#define	SetPgmTitle(pBuf)							_SysSetPgmTitle(SYS_SETPGMTITLE, pBuf)
#define	GetAppDataPath(pBuf)						_SysGetAppDataPath(SYS_GETAPPDATAPATH, pBuf)
#define	GetPgmPath(pBuf)							_SysGetPgmPath(SYS_GETPGMPATH, pBuf)
#define	GetOSversion(pBuf)							_SysGetOSversion(SYS_GETOSVERSION, pBuf)
#define	GetCtl(pBuf)								_SysGetCtl(SYS_GETCTL, pBuf)
#define	SetCursor(curstype)							_SysSetCursor(SYS_SETCURSOR, curstype)
#define	DoHTML(str)									_SysDoHTML(SYS_DOHTML, str);
#define	GetMaxScreen()								_SysGetMaxScreen(SYS_GETMAXSCREEN)
// NOTE: GetKeyStatus is ONLY to be passed MK_SHIFT, MK_CTL, or MK_ALT
#define	GetKeyStatus(k)								_SysGetKeyStatus(SYS_GETKEYSTATUS, k)

#define	iniRead(fname)								_SysiniRead(SYS_INIREAD, fname)
#define	iniWrite(fname)								_SysiniWrite(SYS_INIWRITE, fname)
#define	iniClose()									_SysiniClose(SYS_INICLOSE)
#define	iniFindSec(pSecName)						_SysiniFindSec(SYS_INIFINDSEC, pSecName)
#define	iniFindVar(pVarName)						_SysiniFindVar(SYS_INIFINDVAR, pVarName)
#define	iniGetNum(pNum)								_SysiniGetNum(SYS_INIGETNUM, pNum)
#define	iniGetStr(pStr, maxlen)						_SysiniGetStr(SYS_INIGETSTR, pStr, maxlen)
#define	iniSetSec(pSecName)							_SysiniSetSec(SYS_INISETSEC, pSecName)
#define	iniSetVar(pVarName, pVarVal)				_SysiniSetVar(SYS_INISETVAR, pVarName, pVarVal)
#define	iniDelSec()									_SysiniDelSec(SYS_INIDELSEC)
#define	iniDelVar(pVarName)							_SysiniDelVar(SYS_INIDELVAR, pVarName)
#define	iniNextSec(pBuf, maxlen)					_SysiniNextSec(SYS_ININEXTSEC, pBuf, maxlen)
#define	iniDelAll()									_SysiniDelAll(SYS_INIDELALL)
#define	iniIsEndVar()								_SysiniIsEndVar(SYS_INIENDVAR)
#define	iniAddVarNum(pVarName, val, format)			_SysiniAddVarNum(SYS_INIADDVARNUM, pVarName, val, format)
#define	iniAddVarStr(pVarName, pStr)				_SysiniAddVarStr(SYS_INIADDVARSTR, pVarName, pStr)
#define	iniNextVar(pBuf, maxlen)					_SysiniNextVar(SYS_ININEXTVAR, pBuf, maxlen)

//====================================================================================
// HEAP MANAGEMENT
//====================================================================================
typedef DWORD HEAP;

DWORD _SysHeapAlloc(DWORD cmd, DWORD siz);
void _SysHeapFree(DWORD cmd, HEAP h);
//void *_SysHeapLock(DWORD cmd, HEAP h);
//void _SysHeapUnlock(DWORD cmd, HEAP h);
void *_SysHeapPtr(DWORD cmd, HEAP h);

#define	HeapAlloc(siz)			_SysHeapAlloc(SYS_HEAPALLOC, siz)
#if DEBUG
	#define	HeapFree(h)				(*((DWORD*)0x01F0)=*((DWORD*)0x088), _SysHeapFree(SYS_HEAPFREE, h))
	#define	HeapPtr(h)				(*((DWORD*)0x01F8)=*((DWORD*)0x088), _SysHeapPtr(SYS_HEAPPTR, h))
#else
	#define	HeapFree(h)				_SysHeapFree(SYS_HEAPFREE, h)
	#define	HeapPtr(h)				_SysHeapPtr(SYS_HEAPPTR, h)
#endif
//#define	HeapLock(h)				_SysHeapLock(SYS_HEAPLOCK, h)
//#define	HeapUnlock(h)			_SysHeapUnlock(SYS_HEAPUNLOCK, h)

//====================================================================================
//====================================================================================
// WINDOWING / USER INTERFACE SYSTEM
//====================================================================================
//====================================================================================

#define	CLR_NONE		0x00000000
#define	CLR_RED			0xFF0000FF
#define	CLR_GREEN		0xFF00FF00
#define	CLR_BLUE		0xFFFF0000
#define	CLR_YELLOW		0xFF00FFFF
#define	CLR_CYAN		0xFFFFFF00
#define	CLR_MAGENTA		0xFFFF00FF
#define	CLR_WHITE		0xFFFFFFFF
#define	CLR_BLACK		0xFF000000
#define	CLR_LTGRAY		0xFFC0C0C0
#define	CLR_MDGRAY		0xFF808080
#define	CLR_DKGRAY		0xFF404040
#define	CLR_TEAL		0xFF787800

// - - - - - - - - - - - - - - 

#define	ClrTrans1			0
#define	ClrTrans2			1

#define	ClrMSGfillNORM		2
#define	ClrMSGedgeNORM		3
#define	ClrMSGtextNORM		4
#define	ClrMSGfillATTN		5
#define	ClrMSGedgeATTN		6
#define	ClrMSGtextATTN		7
#define	ClrMSGfillPROG		8

#define	ClrDLGfillTITLE		9
#define	ClrDLGtextTITLE		10

#define	ClrGRPtext			11

#define	ClrTEXTfillNORM		12
#define	ClrTEXTtextNORM		13
#define	ClrTEXTfillHOVER	14
#define	ClrTEXTtextHOVER	15
#define	ClrTEXTfillFOCUS	16
#define	ClrTEXTtextFOCUS	17
#define	ClrTEXTfillCUR		18
#define	ClrTEXTtextCUR		19
#define	ClrTEXTtextDISABLED	20
#define	ClrEDITfillNORM		21
#define	ClrEDITtextNORM		22

#define	ClrEDITcursor		23

#define	ClrMOODname			24

#define	ClrHelpText			25
#define	ClrHelpLink			26
#define	ClrHelpTitle		27
#define	ClrHelpSubTitle		28
#define	ClrHelpBack			29

#define	ClrButtBarFrame		30

#define	ClrPuzzNumFillU		31
#define	ClrPuzzNumTextU		32
#define	ClrPuzzNumFillS		33
#define	ClrPuzzNumTextS		34
#define	ClrPuzzNumEdge		35

//=================================================================
// Locations of things in the UI image in UIBMUs
#define	X_CREDITS		8
#define	Y_CREDITS		0
//#define	X_MENUBUTTF		19
//#define	Y_MENUBUTTF		13
//#define	X_MENUBUTTH		13
//#define	Y_MENUBUTTH		13
#define	X_MENUBUTT		0
#define	Y_MENUBUTT		39
#define	X_VSCROLLBAR	6
#define	Y_VSCROLLBAR	24
#define	X_HSCROLLBAR	15
#define	Y_HSCROLLBAR	24
#define	X_HSCROLLBOX	0
#define	Y_HSCROLLBOX	24
#define	X_VSCROLLBOX	0
#define	Y_VSCROLLBOX	33
#define	X_MOODLINE		0
#define	Y_MOODLINE		2
#define	X_SEPLINE		0
#define	Y_SEPLINE		1
#define	X_CHECKBOX		0
#define	Y_CHECKBOX		4
#define	X_RADIOBUTT		0
#define	Y_RADIOBUTT		8
#define	X_RADIOGREEN	1
#define	Y_RADIOGREEN	13
#define	X_GROUPBOX		0
#define	Y_GROUPBOX		12
#define	X_BUTTON		6
#define	Y_BUTTON		12
#define	X_DIALOG		0
#define	Y_DIALOG		18
#define	X_LISTBOX		18
#define	Y_LISTBOX		33
#define	X_INBOX			12
#define	Y_INBOX			33
#define	X_COPY			0
#define	Y_COPY			44
#define	X_PASTE			8
#define	Y_PASTE			44
#define	X_WNDBACK		0
#define	Y_WNDBACK		60
#define	W_WNDBACK		44
#define	H_WNDBACK		24
#define	X_TOOLBACK		44
#define	Y_TOOLBACK		0
#define	W_TOOLBACK		12
#define	H_TOOLBACK		84

//====================================================================================
// MessageBox() button sets
#define	MB_OK				0
#define	MB_OKCANCEL			1
#define	MB_YESNO			2
#define	MB_YESNOCANCEL		3
#define	MB_BUMMER			4
#define	MB_WAYCOOL			5
#define	MB_WHATEVER			6
#define	MB_YESNOEXIT		7
#define	MB_NOTIFY			8	// i.e., NO buttons

//====================================================================================
#define	MAX_ZSTACK		1024

//====================================================================================
// WND TYPES
#define	WT_WND			0
#define	WT_WNDSCROLL	1
#define	WT_HBAR			2
#define	WT_VBAR			3
#define	WT_BUTTONTEXT	4
#define	WT_BUTTONBM		5
#define	WT_CHECKBOX		6
#define	WT_RADIOBUTTON	7
#define	WT_TEXT			8
#define	WT_EDITLINE		9
#define	WT_EDITMULTI	10
#define	WT_LISTBOX		11
#define	WT_GROUPBOX		12
#define	WT_SCROLLBAR	13
#define	WT_CUSTOM		14

// WND FLAGS
#define	WF_SHOW			0x00000001	// set if visible
#define	WF_DLG			0x00000002	// set if window is a DIALOG
#define	WF_NSC			0x00000004	// (WF_NotifyStringChange) set if window needs to update a string when STR_*.TXT is reloaded
#define	WF_WNDSCROLL	0x00000008	// set if window type is (or derives from) WT_WNDSCROLL
//#define	WF_unused	0x00000010
//#define	WF_unused	0x00000020
//#define	WF_unused	0x00000040
//#define	WF_unused	0x00000080
#define	CF_DEFAULT		0x00000100	// ctl is the default pushbutton for ENTER in a dialog
#define	CF_HELD_DOWN	0x00000200	// ctl (buttons) is in the checked state temporarily by holding down with the LEFT mouse button
#define	CF_CHECKED		0x00000400	// ctl is checked or in the "on" state
#define	CF_DOWN			0x00000600	// either CF_HELD_DOWN or CF_CHECKED is true
#define	CF_DISABLED		0x00000800	// ctl is disabled, cannot be activated until enabled
#define	CF_FOCUS		0x00001000	// ctl gets focus when dialog is created
#define	CF_STR			0x00002000	// ctl has string address rather than blk#/str# pair
#define	CF_TABSTOP		0x00004000	// tab'ing through a dialog will stop on this control
#define	CF_GROUP		0x00008000	// start of a group of ctls that are grouped together for tab'ing and cursoring
#define	CF_OPAQUE		0x00010000	// TEXT controls that get updated need background to be erased when redrawn
#define	CF_LEFT			0x00000000	// set nothing for left-justified TEXT control
#define	CF_CENTER		0x00020000	// center TEXT control
#define	CF_RIGHT		0x00040000	// right-justify TEXT control (this bit is shared with CF_HORSCROLL)
#define	CF_TOP			0x00080000	// top-justify EDITLINE control
#define	CF_BOTTOM		0x00100000	// bottom-justify EDITLINE control

#define	CF_FIXED		0x00200000	// for controls with text, use FIXED-point font
#define	CF_VERSCROLL	0x00200000	// ctl is a VERTICAL scrollbar (this bit is shared with CF_CENTER) (if not set, ctl is a HORIZONTAL scrollbar)

#define	CF_BOLD			0x00400000	// for controls with text, use BOLD font
#define	CF_TRACKING		0x00400000	// ctl is a scrollbar who's scrollbox is being dragged (this bit is shared with CF_RIGHT)

#define	CF_SORT			0x00800000	// ctl is a listbox who's entries are sorted case-sensitively-alphabetically
#define	CF_VSTATE		0x00800000	// ctl is a scrollbar who's UIBM normal/hover/focus images are arranged vertically instead of horizontally

#define	CF_SERIF		0x01000000	// for controls with text, use SERIF font

#define	CF_COLOR		0x02000000	// for CHECKBOX controls that have a "color sample patch" in front
#define	CF_EDITNUM		0x02000000	// for EDITLINE controls, set if the edit text is a NUMBER rather than just plain text
#define	CF_HEX			0x04000000	// for EDITLINE controls, when CF_EDITNUM is set, if CF_HEX=0, the number is decimal format, else it's hex format

#define	CF_ITALIC		0x08000000	// for controls with text, use ITALIC font
#define	CF_SELFDRAWN	0x08000000	// set if subclassed listbox wants to draw items itself (shared with CF_MINUSARROW)

#define	CF_UNDERLINE	0x10000000	// for controls with text, UNDERLINE text
#define	CF_DBLWIDE		0x10000000	// if ButtonBM is double-wide

#define	TF_BREAK_CHAR	0x20000000	// only set during call to DrawTextBM(), breaks text on a character (at right edge)
#define	MEF_TABS		0x20000000	// MultEdit controls that accept TAB key and turn into a TAB character

#define	TF_BREAK_SPACE	0x40000000	// only set during call to DrawTextBM(), breaks text on a SPACE character (at right edge)
#define	MEF_NODRAW		0x40000000	// MultEdit control doesn't redraw screen when this is set
#define	LBF_SORTCOLNUMS	0x40000000

#define	TF_GET_WIDTH	0x80000000	// only set during call to DrawTextBM(), doesn't draw, just returns width of what WOULD be drawn
#define	MEF_CHANGED		0x80000000	// MultEdit control - set when buffer is changed
#define	LBF_SORT_REV	0x80000000

#define	TF_CLEAR		0x1FFFFFFF	// mask for clearing TF_BREAK_CHAR, TF_BREAK_SPACE, TF_GET_WIDTH bits

//--- GENERAL MESSAGES (ALL WINDOWS)
#define WM_CREATE			0	// this window has just been created
#define WM_PAINT			1	// this window should draw itself
#define WM_DESTROY			2	// this window is being destroyed, clean up
#define WM_PARENTSIZED		3	// parent window has been resized
#define WM_PTRMOVE			4	// mouse ptr has moved
#define WM_LBUTTDN			5	// LEFT button has been pressed
#define WM_MBUTTDN			6	// MIDDLE button has been pressed
#define WM_RBUTTDN			7	// RIGHT button has been pressed
#define WM_BBUTTDN			8	// BACK button has been pressed
#define WM_FBUTTDN			9	// FWD button has bee pressed
#define WM_LBUTTUP			10	// LEFT button has been released
#define WM_MBUTTUP			11	// MIDDLE button has been released
#define WM_RBUTTUP			12	// RIGHT button has been released
#define WM_BBUTTUP			13	// BACK button has been released
#define WM_FBUTTUP			14	// FWD button has been released
#define WM_KEYDN			15	// key has been pressed
#define WM_KEYUP			16	// key has been released
#define WM_KILLFOCUS		17	// this window has lost the focus
#define WM_SETFOCUS			18	// focus has been set to this window
#define WM_CHAR				19	// alphanumeric character entered
#define WM_LBUTTDBLCLICK	20	// left-button clicked twice close together
#define WM_SHOW				21	// show or hide window
#define WM_DRAWITEM			22	// draw specified sub-part of window (list box item)
#define WM_DRAWBACKGROUND	23	// draw background of window
#define WM_ACTIVATE			24	// window should take action
#define WM_WHEEL			25	// mouse wheel has turned
#define	WM_GETCOLORS		26	// get colors for drawing a window (window-type specific, arg1 points to a DWORD[])
#define	WM_GETCHILDCOLOR	27	// provide color for child control (arg1=child's HWND, arg2=child's ID, arg3=(DWORD*)ptr to where to store color)
#define	WM_NOTIFY			28	// WND(arg1) has sent (arg2) notification (see WND_ defines)
#define	WM_COMMAND			29	// arg1==WC_ command
#define	WM_HOVER			30	// mouse entered (arg1!=0) or left (arg1==0) this window
#define	WM_HWHEEL			31	// mouse wheel has been tilted left (arg3<0) or right (arg3>0)


//--- SPECIAL MESSAGES (MAIN WND ONLY)
#define WM_MOVE				64	// move the window
#define	WM_QUERY_QUIT		65	// asking permission to quit program
#define	WM_SCREEN_CHANGE	66	// screen resolution, etc, has changed
#define	WM_WINDOW_MINIMIZE	67	// window has been minimized or restored
#define	WM_TIMER			68	// timer has gone off
#define	WM_MUSIC_DONE		69	// music file has finished playing

//---- LIST BOX MESSAGES
#define	LBM_ADDSTRING		96	// add a string to the list box
#define	LBM_DELSTRING		97	// delete a string from the list box
#define	LBM_DELITEM			98	// delete an item # from the list box
#define	LBM_GETCOUNT		99	// get count of # of items in list box
#define	LBM_GETCURSEL		100	// get item# of currently selected item
#define	LBM_SETCURSEL		101	// select item# in list box
#define	LBM_SELSTRING		102	// select string in list box
#define	LBM_SELVAL1			103	// select first item in list box with val1==arg1
#define	LBM_GETITEM			104	// optionally (according to arg3) get arg1 item's val1, val2, and text to buffer pointed to by arg2.  Returns 0 if fail, 1 if succeeded.  If failed, buffer is set to 0's.
// NOTE: for LBM_GETITEM, arg3 is flags that determine WHICH values get written to the arg2 buffer
//	If val1 is to be written, it's written first.
//	If val2 is to be written, it's written next.
//	If text is to be written, it's written last.
//	The following #define's are used (and possibly OR'd or ADDed together) for arg3
	#define	LBGI_VAL1	1
	#define	LBGI_VAL2	2
	#define	LBGI_TEXT	4
#define	LBM_GETCOLORS		105	// get background & foreground colors for column of text in list box
// NOTE: arg1 is formatted as 8 nybbles fcccnnnn where:
//	the 0x80000000 bit of 'f' is set if the listbox has the focus
//	the 0x40000000 bit of 'f' is set if this item is the current selection in the list box
//	the ccc nybbles comprise the column# (starts at 0, is incremented each time a \t is encountered)
//	the nnnn nybbles comprise the item# within the listbox (maximum of decimal 65535 items in list box)
#define	LBM_GETCOLWIDTH		106
// NOTE: arg1 is the current column#.  Return the # of pixels in the width of this column.
//	Usually, you will measure the column width in terms of pWnd->unit, so the should return (width*pWnd->unit)
#define	LBM_SORTCOL			107	// resort the listbox starting in column arg1
#define	LBM_FINDSTRING		108	// find string *arg1, return index# or -1 if not found
#define	LBM_SELVAL12		109	// select first item whose val1==arg1 AND val2==arg2
#define	LBM_FINDVAL1		110	// return item# of first item whose val1==arg1

// EDIT LINE MESSAGES
#define	ELM_LIMITTEXT		112	// set text length limit to arg1
#define	ELM_SETTEXT			113	// set text buffer contents to (char *)arg1 (re-allocate buffer to make it bigger if necessary)
#define	ELM_GETTEXT			114	// copy text from edit buffer to (char *)arg1
#define	ELM_GETNUM			115	// convert text to numeric value and return it
#define	ELM_SETNUM			116	// set edit field to text value of arg1

// HELP MESSAGES
#define	HM_HYPERJUMP		128	// jump to pid arg1 and mid arg2

// RADIO BUTTON MESSAGES
#define	RBM_GETSELINDEX		132	// for Group containing THIS radiobutton, return the index (0-based) of the selected radio button

// MULT EDIT MESSAGES
#define	MEM_INSERTTEXT		136	// insert text pointed to by (BYTE *)arg1 at pC->caret
#define	MEM_GETTEXT			137	// copy edit buffer into buffer pointed to by (BYTE *)arg1, return # bytes placed in buffer
#define	MEM_SETFONT			138	// set the font used by the MultEdit control to arg1, FH to arg2
#define	MEM_LIMITTEXT		139	// set text length limit to arg1
#define	MEM_PASTE			140	// paste arg1:arg2 (memblk:offset) and len of arg3 to current cursor pos
#define	MEM_SETREDRAW		141	// set redraw flag to TRUE or FALSE
#define	MEM_SETMODIFY		142	// set/clear MEF_CHANGED flag
#define	MEM_GETMODIFY		143	// get MEF_CHANGED flag
#define	MEM_FINDSTRING		144	// case-insensitive FindString (selects found string)
#define	MEM_GETLINE			145	// get line arg1 into buffer arg2, max length of arg3
#define	MEM_GETLINECOUNT	146	// return NumLines
#define	MEM_MOVECURSOR		147	// position cursor (and view window if necessary) to line arg1, char offset arg2

// COLOR PICKER MESSAGES
#define	CPM_SETCOLOR_RGB	160	// arg1 = RGB color
#define	CPM_SETCOLOR_HSL	161	// arg1 = HSL color
#define	CPM_GETCOLOR_RGB	162	// return current RGB color
#define	CPM_GETCOLOR_HSL	163	// return current HSL color
#define	CPM_INIT			164	// create edit controls and any other init
#define	CPM_SETHEX			165	// arg1 = 0 for DEC, arg1 != 0 for HEX
#define	CPM_GETHEX			166	// returns 0 for DEC, !0 for HEX

// CHECKBOX MESSAGES
#define	CBM_SETCOLOR_RGB	170	// arg1 = RGB color
#define	CBM_SETCOLOR_HSL	171	// arg1 = HSL color
#define	CBM_GETCOLOR_RGB	172	// return current RGB color
#define	CBM_GETCOLOR_HSL	173	// return current HSL color

// game-specific messages


// END OF KINT WINDOW SYSTEM MESSSAGES



//--- WM_COMMAND WC_ sub-commands
#define	WC_SETSURFSIZE		0	// TO ScrollWnd: set imaginary size of a scrolling window (surfW=arg2, surfH=arg3)
#define	WC_SETPPU			1	// TO ScrollWnd: set pixels-per-unit in X and Y directions (ppuX=arg2, ppuY=arg3)
#define	WC_SETSCROLLPOS		2	// TO ScrollWnd: set scrollbar position X and Y (scrollX=arg2 and scrollY=arg3)
#define	WC_SETSCROLLBARS	3	// TO ScrollWnd: scroll bars need scrollboxes sized and positioned
#define	WC_GETSCROLLINFO	4	// TO Scrollbar: store SBINFO in structure pointed to by arg2
#define	WC_SETSCROLLINFO	5	// TO Scrollbar: load SBINFO from structure pointed to by arg2 (NEVER set TrackPos), AND update scrollbox size and position
#define	WC_GETSCROLLID		6	// TO ScrollWnd: return the ID to be used for arg2==0:Vscrollbar, arg2==1:Hscrollbar
#define	WC_SETPTRSTR		7	// TO Button: set text label ptr 'text' to arg2 (char *) and set CF_STR
#define	WC_SETBLKSTR		8	// TO Button: set text label ptr 'text' to arg2 ((block#<<16)+string#) and clear CF_STR
#define	WC_MOVE				9	// ALL wnds: arg2=x, arg3=y
#define	WC_ZTOP				10	// ALL wnds: move the wnd to the TOP of the ZORDER


//--- WM_NOTIFY SUB-MESSAGES
#define	WN_MODIFIED			1	// window has been modified
#define	WN_CHILDSETFOCUS	2	// dialog's child window has received focus
#define	WN_CHILDKILLFOCUS	3	// dialog's child window has lost focus
#define	WN_CHILDDESTROY		4	// dialog's child window has been destroyed
#define	WN_SCROLLTRACK		5	// window's arg2 child scrollbar has had it's scrollbox dragged (arg3 points to SBINFO)
#define	WN_SCROLL			6	// window's arg2 child scrollbar has been scrolled (arg3 points to SBINFO)
#define	WN_SCROLLUPDATE		7	// ScrollWnd's scrollbars have changed, update SBINFOs
#define	WN_SELCHANGED		8	// To PARENT: arg2==(KSCROLLWNDPTR *)'s listbox who's current selection has changed to arg3
#define	WN_SCROLLED			9	// To PARENT: arg2==(KSCROLLWNDPTR *)'s listbox which has been scrolled
#define	WN_CURSORMOVED		10	// To SELF: window's cursor has moved to line==arg2, column==arg3 (0-based)
#define	WN_STRINGS_RELOADED	11	// The STR_*.TXT file has been reloaded, reset any string pointers/buffers

//--- arg3 values for WM_MOVE command message
#define	SWP_NOMOVE			0x00000001
#define	SWP_NOZORDER		0x00000002

//====================================================================================
// arg3 definitions for WM_PARENTSIZED message to mainWnd (arg1=wndW, arg2=wndH)
#define	K_SIZED		0
#define	K_MINIMIZED	1
#define	K_MAXIMIZED	2
#define	K_RESTORED	3

//====================================================================================
// ShowWnd() commands
#define	SW_HIDE			0
#define	SW_SHOW			1

//====================================================================================
// keycode definitions
#define MK_LBUTT	0x0001
#define MK_RBUTT	0x0002
#define MK_SHIFT	0x0004
#define MK_CTL		0x0008
#define MK_MBUTT	0x0010
#define MK_BBUTT	0x0020
#define MK_FBUTT	0x0040
#define MK_ALT		0x8000

#define K_LBUTT		0x01
#define K_RBUTT		0x02
#define K_MBUTT		0x04

#define K_BACK		0x08
#define K_TAB		0x09
#define	K_CLEAR		0x0C
#define K_ENTER		0x0D

#define K_SHIFT		0x10
#define K_CTL		0x11
#define K_ALT		0x12
//#define	K_PAUSE	0x13
//#define	K_CAPITAL	0x14

#define K_ESC		0x1B

#define K_SPACE		0x20
#define K_PGUP		0x21
#define K_PGDN		0x22
#define K_END		0x23
#define K_HOME		0x24
#define K_LFCURS	0x25
#define K_UPCURS	0x26
#define K_RTCURS	0x27
#define K_DNCURS	0x28
//#define	K_SELECT	0x29
//#define	K_PRINT		0x2A
//#define	K_EXECUTE	0x2B
//#define	K_SNAPSHOT	0x2C
#define K_INS		0x2D
#define K_DEL		0x2E
//#define	K_HELP	0x2F

// These are just their ASCII values (uppercase for letters)
// 0-9 are 0x30 - 0x39
// A-Z are 0x41 - 0x5A
#define	K_0			0x30
#define	K_1			0x31
#define	K_2			0x32
#define	K_3			0x33
#define	K_4			0x34
#define	K_5			0x35
#define	K_6			0x36
#define	K_7			0x37
#define	K_8			0x38
#define	K_9			0x39

#define K_A			0x41
#define K_B			0x42
#define K_C			0x43
#define K_D			0x44
#define K_E			0x45
#define K_F			0x46
#define K_G			0x47
#define K_H			0x48
#define K_I			0x49
#define K_J			0x4A
#define K_K			0x4B
#define K_L			0x4C
#define K_M			0x4D
#define K_N			0x4E
#define K_O			0x4F
#define K_P			0x50
#define K_Q			0x51
#define K_R			0x52
#define K_S			0x53
#define K_T			0x54
#define K_U			0x55
#define K_V			0x56
#define K_W			0x57
#define K_X			0x58
#define	K_Y			0x59
#define	K_Z			0x5A

#define	K_LWIN		0x5B
#define	K_RWIN		0x5C
#define	K_APPS		0x5D

#define K_NUMPAD0	0x60
#define K_NUMPAD1	0x61
#define K_NUMPAD2	0x62
#define K_NUMPAD3	0x63
#define K_NUMPAD4	0x64
#define K_NUMPAD5	0x65
#define K_NUMPAD6	0x66
#define K_NUMPAD7	0x67
#define K_NUMPAD8	0x68
#define K_NUMPAD9	0x69
#define K_MULTIPLY	0x6A
#define K_NUMPADADD	0x6B
#define K_SEPARATOR	0x6C
#define K_NUMPADSUB	0x6D
#define K_DECIMAL	0x6E
#define K_NUMPADDIV	0x6F
#define K_F1		0x70
#define K_F2		0x71
#define K_F3		0x72
#define K_F4		0x73
#define K_F5		0x74
#define K_F6		0x75
#define K_F7		0x76
#define K_F8		0x77
#define K_F9		0x78
#define K_F10		0x79
#define K_F11		0x7A
#define K_F12		0x7B
#define K_F13		0x7C
#define K_F14		0x7D
#define K_F15		0x7E
#define K_F16		0x7F
#define K_F17		0x80
#define K_F18		0x81
#define K_F19		0x82
#define K_F20		0x83
#define K_F21		0x84
#define K_F22		0x85
#define K_F23		0x86
#define K_F24		0x87

#define K_NUMLOCK	0x90
#define K_SCROLL	0x91

#define	K_SEMICOLON		0xBA	// ;: key
#define K_EQPLUS		0xBB	// =+ key
#define	K_COMMA			0xBC	// ,<
#define K_UNDERSUB		0xBD	// -_ key
#define	K_PERIOD		0xBE	// .>
#define K_SLASH			0xBF	// /? key
#define	K_TILDE			0xC0	// `~
#define	K_BACKSLASH		0xDC	// \  key
#define	K_OPENBRACKET	0xDB	// [{
#define	K_BACKSLASH		0xDC	// \|
#define	K_CLOSEBRACKET	0xDD	// ]}

#define K_BACKBUTT  0x8001	// mouse BACK button
#define K_FWDBUTT   0x8002	// mouse FWD button
#define K_WHEEL		0x8003	// mouse wheel
#define	K_HWHEEL	0x8004	// mouse horizontal wheel

#define	MAXTIMERWNDS	16

#define	REPEAT_FIRST	500	// millisecond wait for first repeat
#define	REPEAT_SECOND	200	// millisecond wait for all subsequent repeats

typedef HEAP	HWND;

// DLGRES defines a dialog layout and the layout of its child controls
typedef struct {
		DWORD	id;
		DWORD	type;
		long	ux;
		long	uy;
		long	uw;
		long	uh;
		DWORD	flags;			// WND flags
		DWORD	(*proc)();
		char	*pText;
		void	*xx;			// xx/yy are optional arguments (src bitmap for BUTTONBM, width of EDITLINE box)
		DWORD	yy;				//
	} DLGRES;

//====================================================================================
// WND
//====================================================================================
typedef struct {
		DWORD	len;		// length of this data structure (set by KWINLIB at allocation)
		DWORD	id;			// unique application provided identification number
		DWORD	type;		// type: WT_
		long	xoff;		// x-offset from Parent's xoff, or absolute x-location if no parent
		long	yoff;		// y-offset from Parent
		long	w;			// width of window
		long	h;			// height of window
		long	uu;			// UU value when window was created or last resized
		DWORD	hBM;		// handle to window bitmap, may be NULL (must always be >=w x >=h of window)
		HWND	prevSib;	// handle to WND of previous sibling (0 if none)
		HWND	nextSib;	// handle to WND of next sibling (0 if none)
		HWND	child1;		// handle to WND of first child window (0 if none)
		HWND	parent;		// handle to WND of parent window (0 if none)
		DWORD	flags;		// WF_ flags, DF_ flags, CF_ flags, flags, flags, flags...
		DWORD	(*proc)();	// pointer to window procedure function
		DLGRES	*pRES;		// pointer to the DLGRES[] for the control
		DWORD	text;		// 0 or text (usually a label of some sort) for control
		//============ end of standard WND parameters ========================
	} WND;

typedef WND	TEXT;
typedef WND GROUPBOX;
typedef WND BUTTONTEXT;
typedef WND RADIOBUTT;

//====================================================================================
// WND
//====================================================================================
typedef struct {
		DWORD	len;		// length of this data structure (set by KWINLIB at allocation)
		DWORD	id;			// unique application provided identification number
		DWORD	type;		// type: WT_
		long	xoff;		// x-offset from Parent's xoff, or absolute x-location if no parent
		long	yoff;		// y-offset from Parent
		long	w;			// width of window
		long	h;			// height of window
		long	uu;			// UU value when window was created or last resized
		DWORD	hBM;		// handle to window bitmap, may be NULL (must always be >=w x >=h of window)
		HWND	prevSib;	// handle to WND of previous sibling (0 if none)
		HWND	nextSib;	// handle to WND of next sibling (0 if none)
		HWND	child1;		// handle to WND of first child window (0 if none)
		HWND	parent;		// handle to WND of parent window (0 if none)
		DWORD	flags;		// WF_ flags, DF_ flags, CF_ flags, flags, flags, flags...
		DWORD	(*proc)();	// pointer to window procedure function
		DLGRES	*pRES;		// pointer to the DLGRES[] for the control
		DWORD	text;		// 0 or text (usually a label of some sort) for control
		//============ end of standard WND parameters ========================
		DWORD	color;		// current HSL color of sample color patch if CF_COLOR is set in flags
	} CHECKBOX;

//====================================================================================
//=== NOTE! All extensions of WND that add parameters to it MUST MUST MUST start with
//=== the EXACT SAME IDENTICAL structure as above, so that all routines can access
//=== these fundamental properties without an additional structure level of indexing
//=== and without worrying what EXACT TYPE of window they're dealing with.
//====================================================================================

//====================================================================================
// WNDSCROLL
//====================================================================================
typedef struct {
		DWORD	len;		// length of this data structure (set by KWINLIB at allocation)
		DWORD	id;			// unique application provided identification number
		DWORD	type;		// type of window
		long	xoff;		// x-offset from Parent's xoff, or absolute x-location if no parent
		long	yoff;		// y-offset from Parent
		long	w;			// width of window
		long	h;			// height of window
		long	uu;			// UU value when window was created or last resized
		DWORD	hBM;		// handle to window bitmap (0 if none)
		// if not NULL, BM width must be >= min(w, surfW*ppuX) and BM height must be >= min(h, surfH*ppuY)
		HWND	prevSib;	// handle to WND of previous sibling (0 if none)
		HWND	nextSib;	// handle to WND of next sibling (0 if none)
		HWND	child1;		// handle to WND of first child window (0 if none)
		HWND	parent;		// handle to WND of parent window (0 if none)
		DWORD	flags;		// WF_ flags, DF_ flags, CF_ flags, flags, flags, flags...
		DWORD	(*proc)();	// pointer to window procedure function
		DLGRES	*pRES;		// pointer to the DLGRES[] for the control
		DWORD	text;		// 0 or text for control
		//============ end of standard WND parameters ========================
		long	clientOffX;	// X offset to client portion of window
		long	clientOffY;	// Y offset to client portion of window
		long	clientW;	// width of client portion of window (minus framing)
		long	clientH;	// height of client portion of window (minus framing)
		long	surfW;		// how wide is the display surface (in 'units'), i.e., how far can it be scrolled
		long	surfH;		// how tall is the display surface (in 'units')
		long	ppuX;		// number of pixels per 'unit' in X dir, the window is always scrolled in 'units'
		long	ppuY;		// number of pixels per 'unit' in Y dir
		long	scrollX;	// how far the window is scrolled left (in 'units')
		long	scrollY;	// how far the window is scrolled up (in 'units')
		HWND	hScrollV;	// handle for vertical scrollbar (0 if none)
		HWND	hScrollH;	// handle for horizontal scrollbar (0 if none)
		//============ end of standard SCROLLWND parameters ========================
	} WNDSCROLL;

//====================================================================================
// SBINFO for ScrollBars
//====================================================================================
typedef struct {
		long	TrackPos;	// position of the scrollbar as the scrollbox is DRAGGED
		long	Pos;		// position of the scrollbar when not being dragged
		long	minPos;		// minimum position value
		long	maxPos;		// maximum position value
		long	PageSize;	// number of position values in one logical page (for pg up/dn operations)
	} SBINFO;

//====================================================================================
// SCROLLBAR
//====================================================================================
typedef struct {
		DWORD	len;		// length of this data structure (set by KWINLIB at allocation)
		DWORD	id;			// unique application provided identification number
		DWORD	type;		// type of window
		long	xoff;		// x-offset from Parent's xoff, or absolute x-location if no parent
		long	yoff;		// y-offset from Parent
		long	w;			// width of window
		long	h;			// height of window
		long	uu;			// UU value when window was created or last resized
		DWORD	hBM;		// handle to window bitmap (0 if none)
		// if not NULL, BM width must be >= min(w, surfW*ppuX) and BM height must be >= min(h, surfH*ppuY)
		HWND	prevSib;	// handle to WND of previous sibling (0 if none)
		HWND	nextSib;	// handle to WND of next sibling (0 if none)
		HWND	child1;		// handle to WND of first child window (0 if none)
		HWND	parent;		// handle to WND of parent window (0 if none)
		DWORD	flags;		// WF_ flags, DF_ flags, CF_ flags, flags, flags, flags...
		DWORD	(*proc)();	// pointer to window procedure function
		DLGRES	*pRES;		// pointer to the DLGRES[] for the control
		DWORD	text;		// 0 or text for control
		//============ end of standard WND parameters ========================
		SBINFO	info;		// SBINRO structure for this scrollbar
		long	BoxLen;		// length of scroll box in pixels
		long	BoxPos;		// pixel-offset of box from minimum position
		long	BoxMax;		// pixel-offset of maximum box position
		long	np;			// SBINFO->maxPos - SBINFO->minPos (number of positions)
		HWND	minButt;	// pointer to lf/up child pushbutton
		HWND	maxButt;	// pointer to rt/dn child pushbutton
		HWND	boxButt;	// pointer to the scrollbox child pushbutton
		DWORD	lastRepeatTime;	// SysTimeMilli count of last time scroll button repeated
		DWORD	RepeatInterval;	// millisecond count between repeats
		long	RepeatDir;		// negative = ScrollUp, positive = ScrollDn
		//============ end of standard SCROLLWND parameters ========================
	} SCROLLBAR;

//====================================================================================
// LISTBOX
//====================================================================================
typedef struct {
		DWORD	len;		// length of this data structure (set by KWINLIB at allocation)
		DWORD	id;			// unique application provided identification number
		DWORD	type;		// type of window
		long	xoff;		// x-offset from Parent's xoff, or absolute x-location if no parent
		long	yoff;		// y-offset from Parent
		long	w;			// width of window
		long	h;			// height of window
		long	uu;			// UU value when window was created or last resized
		DWORD	hBM;		// handle to window bitmap (0 if none)
		// if not NULL, BM width must be >= min(w, surfW*ppuX) and BM height must be >= min(h, surfH*ppuY)
		HWND	prevSib;	// handle to WND of previous sibling (0 if none)
		HWND	nextSib;	// handle to WND of next sibling (0 if none)
		HWND	child1;		// handle to WND of first child window (0 if none)
		HWND	parent;		// handle to WND of parent window (0 if none)
		DWORD	flags;		// WF_ flags, DF_ flags, CF_ flags, flags, flags, flags...
		DWORD	(*proc)();	// pointer to window procedure function
		DLGRES	*pRES;		// pointer to the DLGRES[] for the control
		DWORD	text;		// 0 or text for control
		//============ end of standard WND parameters ========================
		long	clientOffX;	// X offset to client portion of window
		long	clientOffY;	// Y offset to client portion of window
		long	clientW;	// width of client portion of window (minus framing)
		long	clientH;	// height of client portion of window (minus framing)
		long	surfW;		// how wide is the display surface (in 'units'), i.e., how far can it be scrolled
		long	surfH;		// how tall is the display surface (in 'units')
		long	ppuX;		// number of pixels per 'unit' in X dir, the window is always scrolled in 'units'
		long	ppuY;		// number of pixels per 'unit' in Y dir
		long	scrollX;	// how far the window is scrolled left (in 'units')
		long	scrollY;	// how far the window is scrolled up (in 'units')
		HWND	hScrollV;	// handle for vertical scrollbar (0 if none)
		HWND	hScrollH;	// handle for horizontal scrollbar (0 if none)
		//============ end of standard SCROLLWND parameters ========================
		DWORD	count;		// how many items are in listbox
		DWORD	cursel;		// currently selected item
		DWORD	firstitem;	// handle of memory block of first item in listbox
		DWORD	sortcol;	// column to use for sorting items
		DWORD	maxW;		// width of longest item
		DWORD	lastHover;	// item# that was last hovered over
	} LISTBOX;

// LISTBOX ITEM
typedef	struct {
		DWORD	hPrev;	// handle of memblk for prev list item
		DWORD	hNext;	// handle of memblk for next list item
		DWORD	val1;	// used for whatever you want
		DWORD	val2;	// used for whatever you want
		DWORD	sortval;	// if sort col==numeric value this is this item's value
		char	text[4];	// storage for string.  Actual length varies (usually > 4)
	} LISTITEM;

//====================================================================================
//====================================================================================
//====================================================================================
// NOTE NOTE NOTE: When CONTROLs are created, they immediately get a WM_CREATE message,
// but BEWARE: if they (as usual) are a child of a dialog, their parent dialog has
// not yet received ITS WM_CREATE message. So, controls shouldn't expect too much of
// their parent during their own WM_CREATE message!!! The dialog receiveds its WM_CREATE
// message AFTER all of its children have been created.	
//====================================================================================
//====================================================================================
//====================================================================================

//====================================================================================
// BUTTONBM
//====================================================================================
// For BUTTONBM:
//		if( w==h ) then the GBMimgW = GBMimgH = UIBMU5
//		else if( w<h ) then the GBMimgW = UIBMU5 and GBMimgH/GBMimgW = h/w, so GBMimgH = GBMimgW*h/w, or GBMimgH = UIBMU5*h/w
//		else then the GBMimgH = GMBU5 and GBMimgW/GBMimgH = w/h, so GBMimgW = GBMimgH*w/h, or GBMimgW = GMBU5*w/h
// The GBM *MUST* contain 4 images for the button (horizontally arranged): no hover/not depressed, hover/not depressed, no hover/depressed, hover/depressed
//
typedef struct {
		DWORD	len;		// length of this data structure (set by KWINLIB at allocation)
		DWORD	id;			// unique application provided identification number
		DWORD	type;		// type: WT_
		long	xoff;		// x-offset from Parent's xoff, or absolute x-location if no parent
		long	yoff;		// y-offset from Parent
		long	w;			// width of window
		long	h;			// height of window
		long	uu;			// UU value when window was created or last resized
		DWORD	hBM;		// handle to window bitmap, may be NULL (must always be >=w x >=h of window)
		HWND	prevSib;	// handle to WND of previous sibling (0 if none)
		HWND	nextSib;	// handle to WND of next sibling (0 if none)
		HWND	child1;		// handle to WND of first child window (0 if none)
		HWND	parent;		// handle to WND of parent window (0 if none)
		DWORD	flags;		// WF_ flags, DF_ flags, CF_ flags, flags, flags, flags...
		DWORD	(*proc)();	// pointer to window procedure function
		DLGRES	*pRES;		// pointer to the DLGRES[] for the control
		DWORD	text;		// 0 or text (usually a label of some sort) for control
		//============ end of standard WND parameters ========================
		DWORD	UIBMUx, UIBMUy, UIBMUw, UIBMUh;// (x,y) coordinates and (w,h) size, in UIBMUs, of 3 "state images"
	} BUTTONBM;

//====================================================================================
// EDITLINE
//====================================================================================

typedef struct {
		DWORD	len;		// length of this data structure (set by KWINLIB at allocation)
		DWORD	id;			// unique application provided identification number
		DWORD	type;		// type: WT_
		long	xoff;		// x-offset from Parent's xoff, or absolute x-location if no parent
		long	yoff;		// y-offset from Parent
		long	w;			// width of window
		long	h;			// height of window
		long	uu;			// UU value when window was created or last resized
		DWORD	hBM;		// handle to window bitmap, may be NULL (must always be >=w x >=h of window)
		HWND	prevSib;	// handle to WND of previous sibling (0 if none)
		HWND	nextSib;	// handle to WND of next sibling (0 if none)
		HWND	child1;		// handle to WND of first child window (0 if none)
		HWND	parent;		// handle to WND of parent window (0 if none)
		DWORD	flags;		// WF_ flags, DF_ flags, CF_ flags, flags, flags, flags...
		DWORD	(*proc)();	// pointer to window procedure function
		DLGRES	*pRES;		// pointer to the DLGRES[] for the control
		DWORD	text;		// 0 or text (usually a label of some sort) for control
		//============ end of standard WND parameters ========================
		DWORD	hBuf;		// handle of EDIT text buffer (on Heap)
		DWORD	buflen;		// len of edit buffer
		DWORD	textlim;	// # of chars (excluding the terminating NUL) allowed in the edit buffer
		DWORD	winS;		// 0-based index into edit buffer of first visible character in edit line window
		DWORD	selS;		// 0-based index into edit buffer of start of selected text
		DWORD	selE;		// 0-based index into edit buffer of end of selected text
		DWORD	caret;		// 0-based index into edit buffer of position of edit cursor
		DWORD	shiftcaret;	// 0-based index into edit buffer of position where SHIFT went down
		DWORD	lastBlinkTime;	// time of last redraw of ctl (for cursor blinking purposes)
		DWORD	cursorOn;	// non-0 if cursor is drawn, 0 if it's not drawn
		long	ex, ey, ew, eh;	// location & size of actual EDIT box
	} EDITLINE;

//====================================================================================
// DIALOG
//====================================================================================
typedef struct {
		DWORD	len;		// length of this data structure (set by KWINLIB at allocation)
		DWORD	id;			// unique application provided identification number
		DWORD	type;		// type: WT_
		long	xoff;		// x-offset from Parent's xoff, or absolute x-location if no parent
		long	yoff;		// y-offset from Parent
		long	w;			// width of window
		long	h;			// height of window
		long	uu;			// UU value when window was created or last resized
		DWORD	hBM;		// handle to window bitmap, may be NULL (must always be >=w x >=h of window)
		HWND	prevSib;	// handle to WND of previous sibling (0 if none)
		HWND	nextSib;	// handle to WND of next sibling (0 if none)
		HWND	child1;		// handle to WND of first child window (0 if none)
		HWND	parent;		// handle to WND of parent window (0 if none)
		DWORD	flags;		// WF_ flags, DF_ flags, CF_ flags, flags, flags, flags...
		DWORD	(*proc)();	// pointer to window procedure function
		DLGRES	*pRES;		// pointer to the DLGRES[] for the dialog
		DWORD	text;		// 0 or text (usually a label of some sort) for control
		//============ end of standard WND parameters ========================
		long	dragDX;		// delta last mouse ptr (x,y) from dlg (x,y)
		long	dragDY;
		DWORD	hLCfocus;	// Last Child that had the focus
	} DIALOG;

//====================================================================================
// COLORPICKER
//====================================================================================
// NOTE: The COLORPICKER is a FIXED SIZE control.
// Any color picker created MUST use CP_W and CP_H as their w/h!
typedef struct {
		DWORD	len;		// length of this data structure (set by KWINLIB at allocation)
		DWORD	id;			// unique application provided identification number
		DWORD	type;		// type of window
		long	xoff;		// x-offset from Parent's xoff, or absolute x-location if no parent
		long	yoff;		// y-offset from Parent
		long	w;			// width of window
		long	h;			// height of window
		long	uu;			// UU value when window was created or last resized
		DWORD	hBM;		// handle to window bitmap (0 if none)
		// if not NULL, BM width must be >= min(w, surfW*ppuX) and BM height must be >= min(h, surfH*ppuY)
		HWND	prevSib;	// handle to WND of previous sibling (0 if none)
		HWND	nextSib;	// handle to WND of next sibling (0 if none)
		HWND	child1;		// handle to WND of first child window (0 if none)
		HWND	parent;		// handle to WND of parent window (0 if none)
		DWORD	flags;		// WF_ flags, DF_ flags, CF_ flags, flags, flags, flags...
		DWORD	(*proc)();	// pointer to window procedure function
		DLGRES	*pRES;		// pointer to the DLGRES[] for the control
		DWORD	text;		// 0 or text for control
		//============ end of standard WND parameters ========================
		DWORD	color;		// current HSL color of Color Picker
		DWORD	clipcolor;	// current HSL color in the Color Picker's clipboard
	} COLORPICKER;

//====================================================================================
// FUNCTIONS
//====================================================================================
long atol(char *buf);
DWORD SendMsg(HWND hWND, DWORD msg, DWORD arg1, DWORD arg2, DWORD arg3);
DWORD DefWndProc(HWND hW, DWORD msg, DWORD arg1, DWORD arg2, DWORD arg3);
HWND CreateWnd(DWORD id, long type, long xoff, long yoff, long w, long h, DWORD flags, HWND hParent, DWORD (*wndproc)());
void NotifyChildren(HWND hW, DWORD msg, DWORD arg1, DWORD arg2, DWORD arg3);
void InvalidateAllWnd(HWND hW);
void InvalidateWnd(HWND hW, long x, long y, long w, long h);
void Invalidate(long x, long y, long w, long h);
void GetRectWnd(HWND hW, RECT *r);
void UpdateInvalid();
BOOL GetInvalidRect(RECT *r);
BOOL IntersectRect(RECT *tr, RECT *sr);
HWND FindChildWndID(HWND hW, DWORD id);
HWND FindWndID(DWORD id);
HWND FindPtChild(HWND hW, long x, long y);
HWND FindPtWnd(long x, long y);
DWORD SendMsgID(DWORD id, DWORD msg, DWORD arg1, DWORD arg2, DWORD arg3);
void SetFocus(HWND hW);
void SetFocusID(DWORD id);
void EnableWndID(DWORD id, DWORD enable);
void ShowWnd(HWND hW, DWORD show);
DWORD DefDlgProc(HWND hW, DWORD msg, DWORD arg1, DWORD arg2, DWORD arg3);
void OnTimer();
HWND AllocateWnd(DWORD id, long type, long xoff, long yoff, long w, long h, DWORD flags, HWND parent, DWORD (*proc)(), long len);
DWORD MessageBox(char *str, DWORD butts);
char *KString(long blknum, long strnum);
void LoadStrings(char *strname);

DWORD DefComButtProc(HWND hW, DWORD msg, DWORD arg1, DWORD arg2, DWORD arg3);
DWORD ButtonProc(HWND hW, DWORD msg, DWORD arg1, DWORD arg2, DWORD arg3);
DWORD ButtonBMProc(HWND hW, DWORD msg, DWORD arg1, DWORD arg2, DWORD arg3);
HWND CreateButton(DWORD id, long x, long y, long w, long h, DWORD flags, DWORD text, HWND parent, DWORD (*wndproc)());
HWND CreateButtonBM(DWORD id, long x, long y, long w, long h, DWORD flags, DWORD text, DWORD UIBMUx, DWORD UIBMUy, DWORD UIBMUw, DWORD UIBMUh, HWND parent, DWORD (*wndproc)());
HWND CreateDialog(DLGRES *pDR, HWND parent);
DWORD DoModal(DLGRES *pDR, HWND parent);
HWND CreateCheckbox(DWORD id, long x, long y, long w, long h, DWORD flags, DWORD text, HWND parent, DWORD (*proc)());
DWORD CheckboxProc(HWND hW, DWORD msg, DWORD arg1, DWORD arg2, DWORD arg3);
DWORD GetCheckID(DWORD id);
void SetCheckID(DWORD id, DWORD checked);
DWORD GroupboxProc(HWND hW, DWORD msg, DWORD arg1, DWORD arg2, DWORD arg3);
DWORD TextProc(HWND hW, DWORD msg, DWORD arg1, DWORD arg2, DWORD arg3);
DWORD RadioButtonProc(HWND hW, DWORD msg, DWORD arg1, DWORD arg2, DWORD arg3);
DWORD EditLineProc(HWND hW, DWORD msg, DWORD arg1, DWORD arg2, DWORD arg3);
DWORD ScrollbarProc(HWND hW, DWORD msg, DWORD arg1, DWORD arg2, DWORD arg3);
DWORD ListBoxProc(HWND hW, DWORD msg, DWORD arg1, DWORD arg2, DWORD arg3);
DWORD ColorPickerProc(HWND hW, DWORD msg, DWORD arg1, DWORD arg2, DWORD arg3);
void CreateMenuCtls(DLGRES *pDR, BOOL setfoc);
void DestroyMenuCtls(DLGRES *pDR);
void StartMusic(char *fname);
void PauseMusic();
void ResumeMusic();
char *GetCtlString(DLGRES *pDR);

long Krand(long lim);	// returns number between 0 and lim-1, inclusive.  lim should be from 1 to 65535.
#define	Ksrand(seed) randseed = seed

//====================================================================================
// EXTERNS
//====================================================================================
extern long	LU;
extern HWND	hFocus;				// handle of window with current focus
extern HWND	hCapture;			// handle of window with mouse capture
extern HWND	hModal;				// handle of window of currently active modal dialog
extern HWND	hMain;				// KINTlib.c requires a wndMain exist SOMEWHERE
extern DWORD	___arg;
extern long	ScreenW, ScreenH;	// pixel w/h of current display, setup by OnBeginStartup() and others
extern long	MaxScreenW, MaxScreenH;		// pixel w/h of virtual display (all monitors together)
extern BOOL	IgnoreMouseUp;
extern DWORD	randseed;
extern long	TEXTYSPACE;
extern long	TEXTH;
extern long		UU, UU2, UU3, UU4, UU5, UU10;	// User Units - pixels per user interface unit, and 4*UU (size of most buttons on screen)
extern QWORD	LineStyleSelected;	// line style used when indicating the selected control
extern DWORD	Colors[64];			// Color pixels from hUIBM
extern long		BlockPaint;
extern HWND		ZStack[MAX_ZSTACK];		// table of WND handles in order of creation, used for search order for mouse clicks (search most recently created (top-most) first)
extern int		Zcnt;					// number of pointers on the ZStack
extern DWORD	MsgBoxCnt;
extern char		KHelpBuf[MAXHELPLEN];			// buffer for HELP.TXT
extern char		*HelpPages[MAXHELPPAGES];	// ptrs to start of parsed help pages
extern char		HelpPageNames[MAXHELPPAGES][32];// name (for links) of each help page
extern long		NumHelpPages;				// number of HelpPages
extern HELPOBJ	*HelpPage;					// HELPOBJ[] array for currently wrapped Help page
extern char		*pAtS[13];				// ptrs to @Sn strings
extern long		HackUU, HackGU;			// UU/GU values from HelpHack.txt (or =0 if none)
extern DWORD	hHackUIBM, hHackGBM;
extern DWORD	hHelpStyle;						// handle of HeapAlloc() to hold the HTML export STYLE block from @S>style@S>
extern char		HelpCredits[256];		// buffer for CREDITS line of current help file (used by HTML export)
extern long		HelpIconX, HelpIconY;			// either HelpIconX==-1 or (HelpIconX*4, HelpIconY*4) is the pixel coordinates in HELP(.PNG) of a 32x32 pixel game icon for HTML export
extern char		*StringsDir, *HelpDir;
extern TABLECOLOR	TableColor[16];
extern long		TableColorCnt;
extern DWORD	resultModalarg2, resultModalarg3;
extern BOOL		HelpChanged;

#if DEBUG
extern char	*KMSGlist[];
#endif
